/*
 *  Copyright (c) 1992-2012 Matisse Software Inc. All Rights Reserved.
 *
 *  This file (in both binary and source code form) is subject to the
 *  Supplemental License Terms governing use, modification and redistribution
 *  of Sample Code accompanying the Matisse(R) software.
 *
 */

// the default Matisse C++ header
#include "matisseCXX.h"
// the local schema header
#include "sql_eg.h"

// namespaces containing Matisse classes
using namespace matisse;
using namespace matisse::reflect;
using namespace matisse::sql;

// namespaces containing the Schema classes
using namespace examples::cxx_examples::chap_9;

// Sample application which shows how to access data via SQL
// queries.  Shows both the traditional form of accessing individual
// attribute values (columns) as well as getting OIDs which map to
// C++ instances.  Database must be loaded before running this
// application; use the supplied sql_eg.sql SQL script.
class SQLExample
{
private:
  MtDatabase db;

public:
  // simple construction to open and connect to database
  SQLExample(const std::string& host, const std::string& dbname, const std::string& dbNamespace)
    : db(host,dbname)
    {
      db.open();
      db.startVersionAccess();
      // Set the SQL CURRENT_NAMESPACE to 'dbNamespace' so there is
      // no need to use the full qualified names to acces the schema objects
      db.setSqlCurrentNamespace(dbNamespace);
      std::cout << "Accessing: " << db << std::endl;
      std::cout << "Sql Default Namespace: " << db.getSqlCurrentNamespace() << std::endl << std::endl;
    }

  // end and close
  ~SQLExample()
    {
      db.endVersionAccess();
      db.close();
    }

  // send traditional query
  void doItPlain()
    {
      // create a statement, execute some sql and iterate through row/cols
      MtStatement stmt(db);
      std::string sqlStr =
        "SELECT e.name, e.id, e.boss.name FROM Employee e WHERE e.id > 2";
      std::cout << "Sql Default Namespace: " << db.getSqlCurrentNamespace() << std::endl;
      std::cout << "Query: " << sqlStr << std::endl;

      MtResultSet res = stmt.executeQuery(sqlStr);

      // list names and types
      // column index is 1-based
      for (unsigned int i = 1; i <= res.getColumnCount(); i++)
      {
        std::cout << "column " << i << " '" << res.getColumnName(i)
                  << "' type is: "
                  << res.getColumnTypeName(i) << std::endl;
      }
      // since we know that we asked for string and integer, we can dump
      //  all the results
      // use the next method to move through rows
      while (res.next())
      {
        std::cout << "Employee name: " << res.getString(1) << ", id: " 
                  << res.getInteger(2) << ", boss : " << res.getString(3)
                  << std::endl;
      }

      // always remember to close the statement when done
      stmt.close();
    }

  // query for objects
  void doItWithObjects()
    {
      // to get a list of objects, use the Ref operator
      MtStatement stmt(db);
      std::string sqlStr = "SELECT Ref(Employee), Ref(boss) FROM Employee WHERE id > 2;";
      std::cout << "Sql Default Namespace: " << db.getSqlCurrentNamespace() << std::endl;
      std::cout << "Query: " << sqlStr << std::endl;

      MtResultSet res = stmt.executeQuery(sqlStr);

      // list names and types
      // column index is 1-based
      for (unsigned int i = 1; i <= res.getColumnCount(); i++)
      {
        std::cout << "column " << i << " '" << res.getColumnName(i)
                  << "' type is: "
                  << res.getColumnTypeName(i) << std::endl;
      }
      // still use the ResultSet, but get the object and coerce to the
      // class we expect (note:  using dynamic_cast<> will throw
      // std::bad_cast if the result type is not acceptable)
      while (res.next())
      {
        // show the class name of the reference
        std::cout << "[class = "
                  << res.getMtObject(1)->getMtClass().getMtName()
                  << "], ";
        // Note:  using dynamic_cast with pointer conversion returns
        // NULL on failure. If you would prefer an exception
        // (std::bad_cast), use object reference to make it an
        // assertion, e.g.:
        // Employee* e = &dynamic_cast<Employee&>(*(res.getMtObject(1)));
        Employee *e = dynamic_cast<Employee *>(res.getMtObject(1));
        Manager *m = dynamic_cast<Manager *>(res.getMtObject(2));
        if (e == NULL)
        {
          std::cout << "Empty Employee from Query!" << std::endl;
          continue;
        }
        std::cout << "Employee name: " << e->getName()
                  << ", id: " << e->getId() << " boss: "
                  << ((m != NULL) ? m->getName() : "[no boss]")
                  << std::endl;
      }
	
      // remember to the close statement when done
      stmt.close();
    }
};

// main is simply a way to get args to the SQLExample class and catch
// any exceptions
int main(int ac, char **av)
{
  if (ac < 3)
  {
    std::cerr << "Usage: " << av[0] << " <host> <db>" << std::endl;
    return -1;
  }

  try {
    SQLExample sqlexample(av[1], av[2], "examples.cxx_examples.chap_9");
    std::cout << "Traditional SQL Query Use..." << std::endl;
    sqlexample.doItPlain();
    std::cout << std::endl;
    std::cout << "Getting Objects from SQL Query..." << std::endl;
    sqlexample.doItWithObjects();
  }
  catch (MtSQLException& sqle)
  {
    std::cerr << "MtSQL Exception: " << sqle << std::endl;
  }
  catch (MtException& mte)
  {
    std::cerr << "Matisse exception: " << mte << std::endl;
  }
  catch (std::exception& e)
  {
    std::cerr << "Standard exception: " << e.what() << std::endl;
  }
  catch (...)
  {
    std::cerr << "Unknown exception" << std::endl;
  }

  return 0;
}
